# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from ..core.variable.evcommon import _ev
from ..localize import _
from ..utils import EPD, EPError, _rand_lst
from . import modcurpl as cp
from . import readtable

cpcache = c.curpl.GetCPCache()


class EUDByteReader:
    """Read byte by byte."""

    def __init__(self):
        c.PushTriggerScope()
        case = (c.Forward(), c.Forward(), c.Forward(), c.Forward())
        self._read = c.RawTrigger(
            nextptr=case[0], actions=c.SetMemory(0x6509B0, c.SetTo, 0)
        )
        self._case = tuple(
            case[i]
            << c.RawTrigger(
                nextptr=readtable._br[i],
                actions=[
                    c.SetNextPtr(self._read, case[(i + 1) % 4]),
                    c.SetMemory(self._read + 348, c.Add, 1) if i == 3 else [],
                ],
            )
            for i in range(4)
        )
        c.PopTriggerScope()

    # -------

    def seekepd(self, epdoffset, *, subp=0, operation=False):
        """Seek EUDByteReader to specific epd player address"""
        operations = [
            (EPD(self._read) + 1, c.SetTo, self._case[subp]),
            (EPD(self._read) + 87, c.SetTo, epdoffset),
        ]
        if operation:
            return operations
        c.NonSeqCompute(operations)

    @c.EUDMethod
    def _seekoffset(self, offset):
        c.f_div(offset, 4, ret=[EPD(self._read) + 87, _ev[3]])
        c.RawTrigger(
            actions=[
                c.SetMemory(self._read + 348, c.Add, -(0x58A364 // 4)),
                c.SetNextPtr(self._read, self._case[0]),
            ]
        )
        for i in range(1, 4):
            c.RawTrigger(
                conditions=_ev[3].Exactly(i),
                actions=c.SetNextPtr(self._read, self._case[i]),
            )

    def seekoffset(self, offset, *, operation=False):
        """Seek EUDByteReader to specific address"""
        if c.IsEUDVariable(offset):
            if operation:
                raise EPError(_("offset is EUDVariable"))
            return self._seekoffset(offset)
        q, r = divmod(offset, 4)
        operations = [
            (EPD(self._read) + 87, c.SetTo, q - (0x58A364 // 4)),
            (EPD(self._read) + 1, c.SetTo, self._case[r]),
        ]
        if operation:
            return operations
        c.NonSeqCompute(operations)

    # -------

    def readbyte(self, *, ret=None):
        """Read byte from current address. EUDByteReader will advance by 1 byte.

        :returns: Read byte
        """
        if ret is None:
            ret = c.EUDVariable()
        try:
            retepd = EPD(ret.getValueAddr())
        except AttributeError:
            retepd = ret

        nexttrg = c.Forward()
        c.RawTrigger(
            nextptr=self._read,
            actions=[
                c.SetNextPtr(readtable.read_end_common, cpcache.GetVTable()),
                c.SetNextPtr(cpcache.GetVTable(), nexttrg),
                cpcache.SetDest(EPD(0x6509B0)),
                c.SetMemory(readtable.read_end_common + 344, c.SetTo, retepd),
            ],
        )
        nexttrg << c.NextTrigger()
        return ret


class EUDByteWriter:
    """Write byte by byte."""

    def __init__(self):
        self._suboffset = c.EUDLightVariable()
        self._write = c.Forward()
        self._return = c.Forward()

        c.PushTriggerScope()
        self._writebyte(0)  # Prevent Forward not initialized error
        # EUDByteWriter._writebyte function has multiple entry points
        self._const = [
            c.RawTrigger(
                conditions=self._suboffset.Exactly(i),
                actions=[
                    c.SetMemory(self._write + 348, c.SetTo, 0),
                    c.SetMemory(self._write + 328, c.SetTo, 0xFF << (8 * i)),
                ],
            )
            for i in range(4)
        ]
        c.SetNextTrigger(self._write)
        c.PopTriggerScope()

    # -------

    def seekepd(self, epdoffset, *, subp=0, operation=False):
        """Seek EUDByteWriter to specific epd player address"""
        operations = [
            (self._suboffset, c.SetTo, subp),
            (EPD(self._write) + 86, c.SetTo, epdoffset),
        ]
        if operation:
            return operations
        c.NonSeqCompute(operations)

    @c.EUDMethod
    def _seekoffset(self, offset):
        # convert offset to epd offset & suboffset
        c.f_div(offset, 4, ret=[EPD(self._write + 344), self._suboffset])
        c.RawTrigger(actions=c.SetMemory(self._write + 344, c.Add, -(0x58A364 // 4)))

    def seekoffset(self, offset, *, operation=False):
        """Seek EUDByteWriter to specific address"""
        if c.IsEUDVariable(offset):
            if operation:
                raise EPError(_("offset is EUDVariable"))
            return self._seekoffset(offset)
        q, r = divmod(offset, 4)
        operations = [
            (EPD(self._write) + 86, c.SetTo, q - (0x58A364 // 4)),
            (self._suboffset, c.SetTo, r),
        ]
        if operation:
            return operations
        c.NonSeqCompute(operations)

    # -------

    @c.EUDMethod
    def _writebyte(self, byte):
        cs.EUDSwitch(self._suboffset)
        if cs.EUDSwitchCase()(0):
            c.VProc(
                byte,
                [
                    byte.SetDest(EPD(self._write + 348)),
                    c.SetMemory(self._write + 328, c.SetTo, 0xFF),
                ],
            )
            cs.EUDBreak()
        for i in range(1, 4):
            if cs.EUDSwitchCase()(i):
                cs.DoActions(
                    c.SetMemory(self._write + 328, c.SetTo, 0xFF << (8 * i))
                )
                for j in _rand_lst(range(8)):
                    c.RawTrigger(
                        conditions=byte.AtLeastX(1, 2**j),
                        actions=c.SetMemory(
                            self._write + 348, c.Add, 2 ** (j + i * 8)
                        ),
                    )
                cs.EUDBreak()

        cs.EUDEndSwitch()

        self._write << c.RawTrigger(
            actions=[
                c.SetDeathsX(0, c.SetTo, 0, 0, 0xFF00),
                self._suboffset.AddNumber(1),
                c.SetMemory(self._write + 348, c.SetTo, 0),
            ]
        )
        c.RawTrigger(
            conditions=[self._suboffset >= 4],
            actions=[
                self._suboffset.SetNumber(0),
                c.SetMemory(self._write + 344, c.Add, 1),
            ],
        )
        self._return << c.NextTrigger()

    def writebyte(self, byte):
        """Write byte to current position.

        Write a byte to current position of EUDByteWriter.
        ByteWriter will advance by 1 byte.
        """
        if c.IsEUDVariable(byte):
            return self._writebyte(byte)

        nextptr = c.Forward()
        c.RawTrigger(
            nextptr=self._const[0],
            actions=[
                *[
                    c.SetMemory(self._const[i] + 348, c.SetTo, byte << (8 * i))
                    for i in range(4)
                ],
                c.SetNextPtr(self._return, nextptr),
            ],
        )
        nextptr << c.NextTrigger()

    def flushdword(self):
        pass


class EUDByteStream:
    """Write and read byte by byte."""

    def __init__(self):
        self._offset = c.EUDVariable()
        self._suboffset = c.EUDVariable()

    # -------

    def seekepd(self, epdoffset):
        """Seek EUDByteWriter to specific epd player address"""
        if c.IsEUDVariable(epdoffset):
            c.VProc(
                epdoffset,
                [
                    *epdoffset.QueueAssignTo(self._offset),
                    self._suboffset.SetNumber(0),
                ],
            )
        else:
            c.RawTrigger(
                actions=[
                    self._offset.SetNumber(epdoffset),
                    self._suboffset.SetNumber(0),
                ]
            )

    @c.EUDMethod
    def seekoffset(self, offset):
        """Seek EUDByteWriter to specific address"""
        # convert offset to epd offset & suboffset
        c.f_div(offset, 4, ret=[self._offset, self._suboffset])
        self._offset += -(0x58A364 // 4)

    def copyto(self, bytestream):
        if isinstance(bytestream, EUDByteStream):
            offset = bytestream._offset
        elif isinstance(bytestream, EUDByteReader):
            offset = EPD(bytestream._read) + 5
        elif isinstance(bytestream, EUDByteWriter):
            offset = EPD(bytestream._write) + 4
        else:
            raise EPError(_("copyto target should be EUDByteReader/Writer/Stream"))
        c.VProc(
            [self._offset, self._suboffset],
            [
                self._offset.SetDest(offset),
                self._suboffset.SetDest(bytestream._suboffset),
            ],
        )

    # -------

    @c.EUDMethod
    def readbyte(self):
        """Read byte from current address. ByteReader will advance by 1 bytes.

        :returns: Read byte
        """
        case = [c.Forward() for _ in range(5)]
        ret = _ev[4]

        c.VProc(
            self._offset,
            [self._offset.SetDest(EPD(0x6509B0)), ret.SetNumber(0)],
        )

        for i in range(4):
            case[i] << c.NextTrigger()
            if i < 3:
                cs.EUDJumpIfNot(self._suboffset == i, case[i + 1])
            for j in _rand_lst(range(8)):
                c.RawTrigger(
                    conditions=c.DeathsX(cp.CP, c.AtLeast, 1, 0, 2 ** (j + 8 * i)),
                    actions=ret.AddNumber(2**j),
                )
            if i < 3:
                c.RawTrigger(nextptr=case[-1], actions=self._suboffset.AddNumber(1))
            else:  # suboffset == 3
                cs.DoActions(self._offset.AddNumber(1), self._suboffset.SetNumber(0))

        case[-1] << c.NextTrigger()
        cp.f_setcurpl2cpcache()
        return ret

    @c.EUDMethod
    def writebyte(self, byte):
        """Write byte to current position.

        Write a byte to current position of EUDByteWriter.
        ByteWriter will advance by 1 byte.
        """
        write = c.Forward()

        c.VProc(
            self._offset,
            [
                self._offset.SetDest(EPD(write) + 4),
                c.SetMemory(write + 20, c.SetTo, 0),
            ],
        )

        cs.EUDSwitch(self._suboffset)
        if cs.EUDSwitchCase()(0):
            c.VProc(
                byte,
                [
                    byte.SetDest(EPD(write) + 5),
                    c.SetMemory(write, c.SetTo, 0xFF),
                ],
            )
            cs.EUDBreak()
        for i in range(1, 4):
            if cs.EUDSwitchCase()(i):
                cs.DoActions(c.SetMemory(write, c.SetTo, 0xFF << (8 * i)))
                for j in _rand_lst(range(8)):
                    c.RawTrigger(
                        conditions=byte.AtLeastX(1, 2**j),
                        actions=c.SetMemory(write + 20, c.Add, 2 ** (j + i * 8)),
                    )
                cs.EUDBreak()

        cs.EUDEndSwitch()

        cs.DoActions(
            self._suboffset.AddNumber(1),
            write << c.SetDeathsX(0, c.SetTo, 0, 0, 0),
        )
        c.RawTrigger(
            conditions=[self._suboffset >= 4],
            actions=[self._offset.AddNumber(1), self._suboffset.SetNumber(0)],
        )

    def flushdword(self):
        pass



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable, Iterator
from typing import TYPE_CHECKING, Any, TypeGuard

from typing_extensions import Self

if TYPE_CHECKING:
    from ...utils import ExprProxy

class ObjAllocator:
    def WriteByte(self, number: int) -> None: ...  # noqa: N802
    def WriteWord(self, number: int) -> None: ...  # noqa: N802
    def WriteDword(self, number) -> None: ...  # noqa: N802
    def WritePack(self, structformat: str, _arglist: list) -> None: ...  # noqa: N802
    def WriteBytes(self, b: bytes) -> None: ...  # noqa: N802
    def WriteSpace(self, ssize: int) -> None: ...  # noqa: N802
    def _write_trigger(self, conditions: int, actions: int) -> None: ...

class PayloadBuffer:
    def WriteByte(self, number: int) -> None: ...  # noqa: N802
    def WriteWord(self, number: int) -> None: ...  # noqa: N802
    def WriteDword(self, number) -> None: ...  # noqa: N802
    def WritePack(self, structformat: str, _arglist: list) -> None: ...  # noqa: N802
    def WriteBytes(self, b: bytes) -> None: ...  # noqa: N802
    def WriteSpace(self, ssize: int) -> None: ...  # noqa: N802
    def _write_trigger(
        self, prevptr, nextptr, conditions: Iterator, actions: Iterator, flags
    ) -> None: ...

class PayloadBuilder:
    def __new__(cls) -> Self: ...
    def alloc_objects(self, found_objects: dict) -> None: ...
    def construct_payload(
        self, found_objects: dict
    ) -> tuple[bytes, list[int], list[int]]: ...
    def register_create_payload_callback(self, f: Callable[[], Any]) -> None: ...
    def register_after_collecting_callback(self, f: Callable[[], Any]) -> None: ...
    def call_callbacks_on_create_payload(self) -> None: ...
    def call_callbacks_after_collecting(self) -> None: ...
    def get_object_addr(self, index: int) -> RlocInt_C: ...

class RlocInt_C:  # noqa: N801
    def __new__(cls, offset: int, rlocmode: int) -> Self: ...
    @property
    def offset(self) -> int: ...
    @property
    def rlocmode(self) -> int: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __add__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __radd__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __sub__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __rsub__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __mul__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __rmul__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __floordiv__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __and__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __rand__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __or__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __ror__(self, rhs: RlocInt_C | int) -> RlocInt_C: ...
    def __invert__(self) -> RlocInt_C: ...
    def __neg__(self) -> RlocInt_C: ...
    def _is_aligned_ptr(self) -> bool: ...
    def _is_ptr(self) -> bool: ...
    def _is_epd(self) -> bool: ...

def RlocInt(offset: int, rlocmode: int) -> RlocInt_C: ...  # noqa: N802
def toRlocInt(x: RlocInt_C | int) -> RlocInt_C: ...  # noqa: N802

class ConstExpr:
    def __new__(cls, baseobj, offset: int = 0, rlocmode: int = 4) -> Self: ...
    # def __init__(self) -> None: ...
    def Evaluate(self) -> RlocInt_C: ...  # noqa: N802
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __add__(self, rhs: int) -> ConstExpr: ...
    def __radd__(self, rhs: int) -> ConstExpr: ...
    def __sub__(self, rhs: int) -> ConstExpr: ...
    def __rsub__(self, rhs: int) -> ConstExpr: ...
    def __mul__(self, rhs: int) -> ConstExpr: ...
    def __rmul__(self, rhs: int) -> ConstExpr: ...
    def __floordiv__(self, rhs: int) -> ConstExpr: ...
    def __mod__(self, rhs: int) -> int: ...
    def __divmod__(self, rhs: int) -> tuple[ConstExpr, int]: ...
    def __neg__(self) -> ConstExpr: ...
    def _is_aligned_ptr(self) -> bool: ...
    def _is_epd(self) -> bool: ...

class Forward(ConstExpr):
    expr: Any
    def __new__(cls) -> Self: ...
    def __lshift__(self, expr: Any) -> Any: ...
    def IsSet(self) -> bool: ...  # noqa: N802
    def Reset(self) -> None: ...  # noqa: N802
    def Evaluate(self) -> RlocInt_C: ...  # noqa: N802
    def __call__(self, *args: Any, **kwds: Any) -> Any: ...
    def __getattr__(self, name: str) -> Any: ...
    def __getitem__(self, name: Any) -> Any: ...
    def __setitem__(self, name: Any, newvalue: Any) -> None: ...

def Evaluate(x: Any) -> RlocInt_C: ...  # noqa: N802
def IsConstExpr(  # noqa: N802
    x: Any,
) -> TypeGuard[int | ConstExpr | ExprProxy[int] | ExprProxy[ConstExpr]]: ...



# Copyright 2014-2019 by trgk, Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from math import ceil

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..collections.eudarray import EUDArray
from ..core.eudfunc.eudf import _EUDPredefineParam
from ..eudlib.utilf.userpl import IsUserCP
from ..localize import _
from ..memio import (
    f_bread_epd,
    f_cunitread_epd,
    f_getcurpl,
    f_setcurpl,
)
from ..memio.rwcommon import br1, cw
from ..scdata import CurrentPlayer, TrgPlayer
from .cpstr import CPString
from .dbstr import DBStringData
from .strcommon import _obfus, epd2s, hptr, ptr2s

prevcp = c.EUDVariable()


@c.EUDFunc
def _pcolor(p):
    p += ut.EPD(0x581D76)
    pcolor = f_bread_epd(p, 2)
    # fmt: off
    pt_arr = EUDArray([
        21, 14, 16, 14, 14, 14, 14, 28, 16, 14, 28, 28, 28, 28, 16, 16,
        21, 21, 21, 21, 21, 21, 21,  6, 21, 21, 21,  8, 17, 17,  8, 17,
        17,  3, 24, 24, 24, 24, 29, 29, 14, 14, 14, 14, 14, 14, 14, 14,
        28, 28, 28, 28, 31, 14, 28, 31, 31, 31, 31, 31, 22, 23, 25, 25,
        21, 21, 21, 21, 21, 14, 21, 21, 21, 16, 30, 29, 29, 29, 30, 29,
        29, 22, 22, 22, 22,  4, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21,
        21, 24, 21, 21, 21, 24, 21,  6, 17, 21, 29, 17,  6, 17,  8,  8,
         8,  8, 27, 17,  3,  7, 14, 16, 14, 14, 14, 14, 14, 28, 28,  2,
        31, 29, 29,  3,  3,  3, 27, 23, 25, 25, 21, 21, 21, 21, 21, 21,
        21, 16, 16, 24, 21, 28, 30, 30, 27, 27, 30,  2, 17, 24, 15, 15,
        14, 16, 14, 16, 16, 14,  3,  3, 21, 21, 21, 21, 21, 21,  6,  8,
         6,  8,  8, 17, 17, 21, 24, 24, 24, 24, 24, 24, 29, 29, 29, 29,
        14, 14, 14, 28, 28, 28, 30,  2, 21, 21, 21, 21, 21, 14, 14, 21,
        21, 21, 21, 21, 21, 21, 21, 29, 21, 21, 21, 21, 21, 21, 21, 21,
        21, 21, 21, 21, 24, 24, 24, 16, 14, 14, 21, 21, 21, 21, 21, 21,
        21, 21, 21, 21, 21, 21, 14, 21, 21, 21, 14, 28, 28, 28,  4,  4,
    ])
    # fmt: on
    pcolor += pt_arr if pt_arr._is_epd() else ut.EPD(pt_arr)
    return pcolor


def PColor(i):  # noqa: N802
    if isinstance(i, TrgPlayer):
        i = c.EncodePlayer(i)
        if not c.IsEUDVariable(i) and i == CurrentPlayer:
            i = prevcp
    return epd2s(_pcolor(i * 2))


def PName(x):  # noqa: N802
    if isinstance(x, TrgPlayer):
        x = c.EncodePlayer(x)
        if not c.IsEUDVariable(x) and x == CurrentPlayer:
            x = prevcp
    return ptr2s(0x57EEEB + 36 * x)


class FixedText:
    def __init__(self, actions_on_exit=[]):
        self._actions_on_exit = ut.FlattenList([actions_on_exit])
        self._txtptr = c.EUDXVariable(ut.EPD(0x640B58), c.SetTo, 0)

    def __enter__(self):
        f_gettextptr(ret=[self._txtptr])
        return self._txtptr

    def __exit__(self, exc_type, exc_value, traceback):
        cs.DoActions(self._actions_on_exit)
        c.VProc(self._txtptr, [])


@c.EUDFunc
def f_gettextptr():
    ret = c.EUDVariable()
    ret << 0
    for i in range(3, -1, -1):
        c.RawTrigger(
            conditions=c.MemoryX(0x640B58, c.AtLeast, 1, 2**i),
            actions=ret.AddNumber(2**i),
        )
    return ret


@c.EUDFunc
def f_getnextchatdst():
    ret = c.EUDVariable()
    ret << ut.EPD(0x640B60)
    for i in range(3, -1, -1):
        c.RawTrigger(
            conditions=c.MemoryX(0x640B58, c.AtLeast, 1, 2**i),
            actions=ret.AddNumber(ceil((2**i) * 54.5)),
        )
    return ret


@c.EUDFunc
def _addstr_cp():
    if cs.EUDInfLoop()():
        b = br1.readbyte()
        cs.EUDBreakIf(b == 0)
        cw.writebyte(b)
    cs.EUDEndInfLoop()

    cw.flushdword()


@c.EUDFunc
def f_cpstr_addstr(src):
    """Print string as string to CurrentPlayer

    :param src: Source address (Not EPD player)
    """
    br1.seekoffset(src)
    _addstr_cp()


@c.EUDFunc
def f_cpstr_addstr_epd(epd):
    """Print string as string to CurrentPlayer

    :param epd: EPD player of Source address
    """
    br1.seekepd(epd)
    _addstr_cp()


@c.EUDFunc
def f_cpstr_adddw(number):
    """Print number as string to CurrentPlayer.

    :param number: DWORD to print
    """
    skipper = [c.Forward() for _ in range(9)]
    ch = [0] * 10

    # Get digits
    for i in range(10):
        number, ch[i] = c.f_div(number, 10)
        if i != 9:
            cs.EUDJumpIf(number == 0, skipper[i])

    # print digits
    for i in range(9, -1, -1):
        if i != 9:
            skipper[i] << c.NextTrigger()
        cw.writebyte(ch[i] + b"0"[0])

    cw.flushdword()


@c.EUDFunc
def f_cpstr_addptr(number):
    """Print number as string to CurrentPlayer.

    :param number: DWORD to print
    """
    digit = [c.EUDLightVariable() for _ in range(8)]
    cs.DoActions(
        *[digit[i].SetNumber(0) for i in range(8)],
        c.SetDeaths(CurrentPlayer, c.SetTo, ut.b2i4(b"0000"), 0),
    )

    def f(x):
        t = x % 16
        q, r = divmod(t, 4)
        return 2 ** (r + 8 * (3 - q))

    for i in range(31, -1, -1):
        c.RawTrigger(
            conditions=number.AtLeastX(1, 2**i),
            actions=[
                digit[i // 4].AddNumber(2 ** (i % 4)),
                c.SetDeaths(CurrentPlayer, c.Add, f(i), 0),
            ],
        )
        if i % 16 == 0:
            for j in range(4):
                c.RawTrigger(
                    conditions=digit[j + 4 * (i // 16)].AtLeast(10),
                    actions=c.SetDeaths(
                        CurrentPlayer,
                        c.Add,
                        (b"A"[0] - b":"[0]) * (256 ** (3 - j)),
                        0,
                    ),
                )
            if i == 16:
                cs.DoActions(
                    *c.AddCurrentPlayer(1),
                    c.SetDeaths(CurrentPlayer, c.SetTo, ut.b2i4(b"0000"), 0),
                )
            else:
                cs.DoActions(*c.AddCurrentPlayer(1))


def f_cpstr_print(*args, EOS=True, encoding="UTF-8"):  # noqa: N803
    """Print multiple string / number to CurrentPlayer.

    :param args: Things to print

    """
    for arg in ut.FlattenIter(args):
        try:
            arg = arg.fmt()
        except AttributeError:
            pass
        x = ut.unProxy(arg)
        if isinstance(x, str):
            x = x.encode(encoding)
        elif isinstance(x, int):
            x = str(x & 0xFFFFFFFF).encode(encoding)
        if isinstance(x, bytes):
            if x == b"":
                continue
            _obfus.cp_print(x)
        elif isinstance(x, CPString):
            x.Display()
        elif isinstance(x, c.Db | DBStringData):
            f_cpstr_addstr_epd(ut.EPD(x))
        elif isinstance(x, ptr2s):
            f_cpstr_addstr(x._value)
        elif isinstance(x, epd2s):
            f_cpstr_addstr_epd(x._value)
        elif isinstance(x, hptr):
            f_cpstr_addptr(x._value)
        elif isinstance(x, c.EUDVariable | c.ConstExpr):
            f_cpstr_adddw(x)
        else:
            e = _("Object with unknown parameter type {} given to {}")
            raise ut.EPError(e.format(arg, "f_cpstr_print"))
    if EOS:
        cs.DoActions(c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0))


create_unit = c.Forward()


@_EUDPredefineParam((ut.EPD(create_unit + 328 + 32 + 16),))
@c.EUDTypedFunc([TrgPlayer])
def f_raise_CCMU(player):  # noqa: N802
    f_cunitread_epd(ut.EPD(0x628438), ret=[ut.EPD(create_unit + 328 + 32 * 2 + 20)])
    create_unit << c.RawTrigger(
        actions=[
            c.SetMemory(0x628438, c.SetTo, 0),
            c.CreateUnit(1, 0, 64, 0),
            c.SetMemory(0x628438, c.SetTo, 0),
        ]
    )


_eprintln_template = c.Forward()
_eprintln_desync = c.Forward()
_eprintln_print = c.Forward()
_eprintln_eos = c.Forward()
_eprintln_end = c.Forward()


def _eprint_init():
    if _eprintln_template.IsSet():
        return
    c.PushTriggerScope()
    _eprintln_template << c.NextTrigger()
    f_raise_CCMU(CurrentPlayer)
    if cs.EUDIf()(_eprintln_desync << IsUserCP()):
        prevcp << f_getcurpl()
        _eprintln_print << c.RawTrigger(
            nextptr=0, actions=c.SetCurrentPlayer(ut.EPD(0x640B60 + 218 * 12))
        )
        _eprintln_eos << c.RawTrigger(
            actions=c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0)
        )
        f_setcurpl(prevcp)
    cs.EUDEndIf()
    _eprintln_end << c.RawTrigger(nextptr=0)
    c.PopTriggerScope()


def eprint_all(*args):
    _eprint_init()

    _print, _next = c.Forward(), c.Forward()
    c.RawTrigger(
        nextptr=_eprintln_template,
        actions=[
            c.SetMemory(_eprintln_template + 380, c.SetTo, 17),  # AllPlayers
            c.SetMemoryX(_eprintln_desync + 12, c.SetTo, 0, 0xFF000000),
            c.SetNextPtr(_eprintln_print, _print),
            c.SetNextPtr(_eprintln_end, _next),
        ],
    )
    _print << c.NextTrigger()
    f_cpstr_print(*args, EOS=False, encoding="UTF-8")
    c.SetNextTrigger(_eprintln_eos)
    _next << c.NextTrigger()


def f_eprintln(*args):
    _eprint_init()

    _print, _next = c.Forward(), c.Forward()
    c.RawTrigger(
        nextptr=_eprintln_template,
        actions=[
            c.SetMemory(
                _eprintln_template + 380,
                c.SetTo,
                c.EncodePlayer(CurrentPlayer),
            ),
            c.SetMemoryX(_eprintln_desync + 12, c.SetTo, 15 << 24, 0xFF000000),
            c.SetNextPtr(_eprintln_print, _print),
            c.SetNextPtr(_eprintln_end, _next),
        ],
    )
    _print << c.NextTrigger()
    f_cpstr_print(*args, EOS=False, encoding="UTF-8")
    c.SetNextTrigger(_eprintln_eos)
    _next << c.NextTrigger()



from helper import *

from eudplib.scdata.offsetmap import (
    ByteMember,
    DwordMember,
    EPDOffsetMap,
    WordMember,
)


class T(EPDOffsetMap):
    a = DwordMember("struct", 0x00)
    b = WordMember("struct", 0x00)
    c = WordMember("struct", 0x02)
    d = DwordMember("struct", 0x04)
    e = ByteMember("struct", 0x07)


@TestInstance
def test_ctypes():
    a = Db(b"\x00\x01\x02\x03\x04\x05\x06\x07")
    # FIXME: with expect_error(TypeError):
    t = T(a)
    t = T(EUDVariable(EPD(a)))
    test_equality(
        "Reading from EPDOffsetMap",
        [t.a, t.b, t.c, t.d, t.e],
        [0x03020100, 0x0100, 0x0302, 0x07060504, 0x07],
    )

    t.a = 0x0D0C0B0A
    t.c = 0x0302
    t.e = 0x11

    test_equality(
        "Writing to EPDOffsetMap",
        [f_dwread_epd(EPD(a) + 0), f_dwread_epd(EPD(a) + 1)],
        [0x03020B0A, 0x11060504],
    )



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: I001
# fmt: off
from ..core import (
    f_div,
    EUDVariable, EUDCreateVariables, SetVariables,
    VProc,
    Db,
    RawTrigger, PreserveTrigger,
    UnitProperty, Disabled,
    # EUDFunc
    EUDFunc, EUDTypedFunc, EUDFuncPtr, EUDTypedFuncPtr,
    EUDMethod, EUDTypedMethod, EUDReturn,
    # EUDStruct
    EUDStruct, EUDVArray, selftype,
    # ConstType
    Iscript,
    Portrait, Icon,
    StatText,
    TrgAIScript, TrgAllyStatus, TrgComparison, TrgCount,
    TrgLocation, TrgModifier, TrgOrder, TrgProperty,
    TrgPropState, TrgResource, TrgScore, TrgString, TrgSwitch,
    TrgSwitchAction, TrgSwitchState,
    # Condition
    Always, Never,
    Accumulate, Bring,
    Command, CommandLeast, CommandLeastAt, CommandMost, CommandMostAt,
    CountdownTimer, ElapsedTime,
    Deaths, DeathsX, Kills,
    Memory, MemoryEPD, MemoryX, MemoryXEPD,
    HighestScore, MostKills, MostResources,
    LowestScore, LeastKills, LeastResources,
    Opponents,
    Score, Switch,
    # Action
    SetCurrentPlayer,
    CenterView, CreateUnit, CreateUnitWithProperties, Comment,
    Defeat, DisplayText, Draw,
    GiveUnits,
    KillUnit, KillUnitAt,
    LeaderBoardComputerPlayers,
    LeaderBoardControl, LeaderBoardGoalControl,
    LeaderBoardControlAt, LeaderBoardGoalControlAt,
    LeaderBoardGreed, LeaderBoardResources, LeaderBoardGoalResources,
    LeaderBoardKills, LeaderBoardGoalKills,
    LeaderBoardScore, LeaderBoardGoalScore,
    MinimapPing,
    ModifyUnitEnergy, ModifyUnitHangarCount, ModifyUnitHitPoints,
    ModifyUnitResourceAmount, ModifyUnitShields,
    MoveLocation, MoveUnit,
    Order,
    MuteUnitSpeech, PauseGame, PauseTimer,
    UnMuteUnitSpeech, UnpauseGame, UnpauseTimer,
    PlayWAV,
    RemoveUnit, RemoveUnitAt,
    RunAIScript, RunAIScriptAt,
    SetAllianceStatus, SetCountdownTimer,
    SetDeaths, SetDeathsX, SetKills,
    SetDoodadState, SetInvincibility,
    SetMemory, SetMemoryEPD, SetMemoryX, SetMemoryXEPD,
    SetMissionObjectives, SetNextScenario,
    SetResources, SetScore, SetSwitch,
    TalkingPortrait, Transmission,
    Victory,
    Wait,
    # TrgResource
    Ore, Gas, OreAndGas,
    # TrgAllyStatus
    Ally, Enemy, AlliedVictory,
    # TrgOrder
    Attack, Move, Patrol,
    # TrgScore
    Units, Buildings, UnitsAndBuildings,
    Razings, KillsAndRazings, Total, Custom,
    # TrgComparison
    AtLeast, AtMost, Exactly,
    # TrgModifier
    Add, Subtract, SetTo,
    # TrgPropState
    Enable, Disable, Toggle,
    # TrgSwitchState
    Set, Cleared,
    # TrgSwitchAction
    Clear, Random,
    # TrgCount
    All,
)
from ..utils import EPD
from ..ctrlstru import (
    DoActions, EUDTernary, EUDLoopRange,
    EUDWhile, EUDWhileNot, EUDEndWhile,
    EUDBreak, EUDContinue, EUDSetContinuePoint,
    EUDContinueIf, EUDContinueIfNot,
    EUDSCAnd, EUDSCOr,
    EUDIf, EUDIfNot, EUDEndIf,
    EUDElseIf, EUDElseIfNot, EUDElse,
    EUDSwitch, EPDSwitch, EUDEndSwitch,
    EUDSwitchCase, EUDSwitchDefault,
)
from ..memio import (
    f_getcurpl, f_setcurpl,
    f_dwread_epd, f_dwepdread_epd, f_epdread_epd,
    f_dwwrite_epd, f_dwadd_epd, f_dwsubtract_epd,
    f_wread_epd, f_wwrite_epd,
    f_bread_epd, f_bwrite_epd,
    f_maskread_epd, f_posread_epd,
    f_dwread, f_dwwrite,
    f_wread, f_wwrite,
    f_bread, f_bwrite,
)
from ..collections import (
    EUDArray, EUDDeque, EUDQueue, EUDStack,
    PVariable, UnitGroup
)
from ..scdata import (
    TrgUnit, Weapon, UnitOrder, CUnit,
    Flingy, Sprite, Image,
    Upgrade, Tech,
    TrgPlayer,
    P1, P2, P3, P4, P5, P6, P7, P8, P9, P10, P11, P12,
    Force1, Force2, Force3, Force4, AllPlayers, CurrentPlayer,
    Allies, Foes, NeutralPlayers, NonAlliedVictoryPlayers,
)
from ..eudlib.locf.locf import (
    f_setloc, f_addloc, f_dilateloc,
    f_getlocTL, f_setloc_epd,
)
from ..eudlib.mathf.atan2 import f_atan2, f_atan2_256
from ..eudlib.mathf.lengthdir import f_lengthdir, f_lengthdir_256
from ..eudlib.mathf.pow import f_pow
from ..eudlib.mathf.sqrt import f_sqrt
from ..eudlib.utilf.binsearch import EUDBinaryMax, EUDBinaryMin
from ..eudlib.utilf.listloop import (
    EUDLoopCUnit, EUDLoopNewCUnit, EUDLoopPlayerCUnit,
)
from ..eudlib.utilf.pexist import EUDLoopPlayer, f_playerexist
from ..eudlib.utilf.random import f_dwrand, f_getseed, f_rand, f_randomize, f_srand
from ..eudlib.utilf.userpl import (
    IsUserCP, f_getuserplayerid,
    DisplayTextAll, PlayWAVAll, CenterViewAll,
    MinimapPingAll, SetMissionObjectivesAll, TalkingPortraitAll,
)
from ..eudlib.wireframe.wireframe import (
    InitialWireframe, Is64BitWireframe,
    SetGrpWire, SetTranWire,  SetWirefram, SetWireframes,
)
from ..string import (
    IsPName, SetPNamef,
    StringBuffer,
    f_println, f_printAt,
    f_printAll, f_printAllAt,
    f_simpleprint,
    f_eprintf, f_settblf,
)

__all__ = [
    "f_div",
    "EUDVariable", "EUDCreateVariables", "SetVariables",
    "VProc",
    "Db",
    "RawTrigger", "PreserveTrigger",
    "UnitProperty", "Disabled",
    # EUDFunc
    "EUDFunc", "EUDTypedFunc", "EUDFuncPtr", "EUDTypedFuncPtr",
    "EUDMethod", "EUDTypedMethod", "EUDReturn",
    # EUDStruct
    "EUDStruct", "EUDVArray", "selftype",
    # ConstType
    "Weapon", "Flingy", "Sprite", "Image", "Iscript",
    "Portrait", "Icon", "Upgrade", "Tech",
    "UnitOrder", "StatText",
    "TrgAIScript", "TrgAllyStatus", "TrgComparison", "TrgCount",
    "TrgLocation", "TrgModifier", "TrgOrder", "TrgPlayer", "TrgProperty",
    "TrgPropState", "TrgResource", "TrgScore", "TrgString", "TrgSwitch",
    "TrgSwitchAction", "TrgSwitchState", "TrgUnit",
    # TrgPlayer
    "P1", "P2", "P3", "P4", "P5", "P6", "P7", "P8", "P9", "P10", "P11", "P12",
    "Force1", "Force2", "Force3", "Force4", "AllPlayers", "CurrentPlayer",
    "Allies", "Foes", "NeutralPlayers", "NonAlliedVictoryPlayers",
    # Condition
    "Always", "Never",
    "Accumulate", "Bring",
    "Command", "CommandLeast", "CommandLeastAt", "CommandMost", "CommandMostAt",
    "CountdownTimer", "ElapsedTime",
    "Deaths", "DeathsX", "Kills",
    "Memory", "MemoryEPD", "MemoryX", "MemoryXEPD",
    "HighestScore", "MostKills", "MostResources",
    "LowestScore", "LeastKills", "LeastResources",
    "Opponents",
    "Score", "Switch",
    # Action
    "SetCurrentPlayer",
    "CenterView", "CreateUnit", "CreateUnitWithProperties", "Comment",
    "Defeat", "DisplayText", "Draw",
    "GiveUnits",
    "KillUnit", "KillUnitAt",
    "LeaderBoardComputerPlayers",
    "LeaderBoardControl", "LeaderBoardGoalControl",
    "LeaderBoardControlAt", "LeaderBoardGoalControlAt",
    "LeaderBoardGreed", "LeaderBoardResources", "LeaderBoardGoalResources",
    "LeaderBoardKills", "LeaderBoardGoalKills",
    "LeaderBoardScore", "LeaderBoardGoalScore",
    "MinimapPing",
    "ModifyUnitEnergy", "ModifyUnitHangarCount", "ModifyUnitHitPoints",
    "ModifyUnitResourceAmount", "ModifyUnitShields",
    "MoveLocation", "MoveUnit",
    "Order",
    "MuteUnitSpeech", "PauseGame", "PauseTimer",
    "UnMuteUnitSpeech", "UnpauseGame", "UnpauseTimer",
    "PlayWAV",
    "RemoveUnit", "RemoveUnitAt",
    "RunAIScript", "RunAIScriptAt",
    "SetAllianceStatus", "SetCountdownTimer",
    "SetDeaths", "SetDeathsX", "SetKills",
    "SetDoodadState", "SetInvincibility",
    "SetMemory", "SetMemoryEPD", "SetMemoryX", "SetMemoryXEPD",
    "SetMissionObjectives", "SetNextScenario",
    "SetResources", "SetScore", "SetSwitch",
    "TalkingPortrait", "Transmission",
    "Victory",
    "Wait",
    # TrgResource
    "Ore", "Gas", "OreAndGas",
    # TrgAllyStatus
    "Ally", "Enemy", "AlliedVictory",
    # TrgOrder
    "Attack", "Move", "Patrol",
    # TrgScore
    "Units", "Buildings", "UnitsAndBuildings",
    "Razings", "KillsAndRazings", "Total", "Custom",
    # TrgComparison
    "AtLeast", "AtMost", "Exactly",
    # TrgModifier
    "Add", "Subtract", "SetTo",
    # TrgPropState
    "Enable", "Disable", "Toggle",
    # TrgSwitchState
    "Set", "Cleared",
    # TrgSwitchAction
    "Clear", "Random",
    # TrgCount
    "All",
    # utils
    "EPD",
    # control structures
    "DoActions", "EUDTernary", "EUDLoopRange",
    "EUDWhile", "EUDWhileNot", "EUDEndWhile",
    "EUDBreak", "EUDContinue", "EUDSetContinuePoint",
    "EUDContinueIf", "EUDContinueIfNot",
    "EUDSCAnd", "EUDSCOr",
    "EUDIf", "EUDIfNot", "EUDEndIf",
    "EUDElseIf", "EUDElseIfNot", "EUDElse",
    "EUDSwitch", "EPDSwitch", "EUDEndSwitch",
    "EUDSwitchCase", "EUDSwitchDefault",
    # memio
    "f_getcurpl", "f_setcurpl",
    "f_dwread_epd", "f_dwepdread_epd", "f_epdread_epd",
    "f_dwwrite_epd", "f_dwadd_epd", "f_dwsubtract_epd",
    "f_wread_epd", "f_wwrite_epd",
    "f_bread_epd", "f_bwrite_epd",
    "f_maskread_epd", "f_posread_epd",
    "f_dwread", "f_dwwrite",
    "f_wread", "f_wwrite",
    "f_bread", "f_bwrite",
    # collections
    "EUDArray", "EUDDeque", "EUDQueue", "EUDStack",
    "PVariable", "UnitGroup",
    # offsetmap
    "CUnit", "EUDLoopCUnit", "EUDLoopNewCUnit", "EUDLoopPlayerCUnit",
    # eudlib
    "f_setloc", "f_addloc", "f_dilateloc",
    "f_getlocTL", "f_setloc_epd",
    "f_atan2", "f_atan2_256",
    "f_lengthdir", "f_lengthdir_256",
    "f_pow",
    "f_sqrt",
    "EUDBinaryMax", "EUDBinaryMin",
    "EUDLoopPlayer", "f_playerexist",
    "f_dwrand", "f_getseed", "f_rand", "f_randomize", "f_srand",
    "IsUserCP", "f_getuserplayerid",
    "DisplayTextAll", "PlayWAVAll", "CenterViewAll",
    "MinimapPingAll", "SetMissionObjectivesAll", "TalkingPortraitAll",
    "InitialWireframe", "Is64BitWireframe",
    "SetGrpWire", "SetTranWire",  "SetWirefram", "SetWireframes",
    # string
    "IsPName", "SetPNamef",
    "StringBuffer",
    "f_println", "f_printAt",
    "f_printAll", "f_printAllAt",
    "f_simpleprint",
    "f_eprintf", "f_settblf",
]



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) import .BGM_eps_test;
# (Line 2) import .test_eps_stattext as stat;
try:
    BGM_eps_test = _RELIMP(".", "BGM_eps_test")
except ImportError:
    from . import BGM_eps_test
# (Line 3) import py_math;
try:
    stat = _RELIMP(".", "test_eps_stattext")
except ImportError:
    from . import test_eps_stattext as stat
import math
# (Line 4) import py_sys;
import sys
# (Line 5) var x = 1 << 4;
x = _TYGV([None], lambda: [_LSH(1,4)])
# (Line 6) EUDOnStart(function () { x += x; });
@EUDFunc
def _lambda1():
    x.__iadd__(x)

EUDOnStart(_lambda1)
# (Line 7) object ParticleBag extends EUDStruct {
# (Line 8) var a;
class ParticleBag(EUDStruct):
    # (Line 9) var b;
    # (Line 10) };
    # (Line 11) object ObjectWithTooManyFields {
    _fields_ = [
        'a',
        'b',
    ]

# (Line 12) var v0, v1, v2, v3, v4, v5, v6, v7, v8, v9;
class ObjectWithTooManyFields(EUDStruct):
    # (Line 13) };
    # (Line 14) object ClassInheritingBigObj extends ObjectWithTooManyFields {
    _fields_ = [
        'v0',
        'v1',
        'v2',
        'v3',
        'v4',
        'v5',
        'v6',
        'v7',
        'v8',
        'v9',
    ]

# (Line 15) var v10;
class ClassInheritingBigObj(ObjectWithTooManyFields):
    # (Line 16) };
    # (Line 18) function test_compatibility() {
    _fields_ = [
        'v10',
    ]

@EUDFunc
def f_test_compatibility():
    # (Line 19) py_exec("from helper import *\n\
    # (Line 22) ");
    exec("from helper import *\nwith expect_eperror():\n    ClassInheritingBigObj.alloc()\n")
    # (Line 23) static var ret = 0;
    ret = _TYSV([None], [0])
    # (Line 24) const empty = Db(i2b4(0));
    empty = Db(i2b4(0))
    # (Line 25) const cond = Forward();
    cond = Forward()
    # (Line 27) py_exec("from helper import *\n\
    # (Line 36) ");
    exec("from helper import *\nwith expect_eperror():\n    Trigger(cond, ret.AddNumber(1 << 0))\nwith expect_eperror():\n    Trigger(empty, ret.AddNumber(1 << 1))\nwith expect_eperror():\n    Trigger(empty + 1, ret.AddNumber(1 << 2))\nwith expect_eperror():\n    SetVariables(ret, -1, EUDVariable(EncodeModifier(SetTo)))\n")
    # (Line 37) py_exec("from helper import *\n\
    # (Line 45) ");
    exec("from helper import *\nwith expect_eperror():\n    SetVariables(EUDVariable(), 1)\nwith expect_eperror():\n    SetVariables(f_dwread_epd(0), 1)\npv = PVariable()\nwith expect_eperror():\n    SetVariables(pv[0], 1)\n")
    # (Line 46) const tc = GetTriggerCounter();
    tc = GetTriggerCounter()
    # (Line 47) var x0 = EUDVariable(0);
    x0 = _TYLV([None], [EUDVariable(0)])
    # (Line 48) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 50) var x1 = list(EUDVariable(0));
    x1 = _TYLV([None], [FlattenList([EUDVariable(0)])])
    # (Line 51) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 53) var x2 = list(list(EUDVariable(0)));
    x2 = _TYLV([None], [FlattenList([FlattenList([EUDVariable(0)])])])
    # (Line 54) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 56) var x3 = ExprProxy(EUDVariable(0));
    x3 = _TYLV([None], [ExprProxy(EUDVariable(0))])
    # (Line 57) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 59) var x4 = ExprProxy(ExprProxy(EUDVariable(0)));
    x4 = _TYLV([None], [ExprProxy(ExprProxy(EUDVariable(0)))])
    # (Line 60) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 62) var x5, y5 = EUDVariable(0), EUDVariable(0);
    x5, y5 = _TYLV([None, None], [EUDVariable(0), EUDVariable(0)])
    # (Line 63) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 65) var x6, y6 = list(EUDVariable(0), EUDVariable(0));
    x6, y6 = _TYLV([None, None], [FlattenList([EUDVariable(0), EUDVariable(0)])])
    # (Line 66) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 68) var x7, y7, z7 = EUDCreateVariables(3);
    x7, y7, z7 = _TYLV([None, None, None], [EUDCreateVariables(3)])
    # (Line 69) ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 71) const vlist = EUDCreateVariables(3);
    vlist = EUDCreateVariables(3)
    # (Line 72) var x8, y8, z8 = vlist;
    x8, y8, z8 = _TYLV([None, None, None], [vlist])
    # (Line 73) ep_assert(tc + 1 == GetTriggerCounter(), "Wrongly elide var copy");
    ep_assert(tc + 1 == GetTriggerCounter(), "Wrongly elide var copy")
    # (Line 75) const tc2 = GetTriggerCounter();
    tc2 = GetTriggerCounter()
    # (Line 76) var x9, y9, z9 = list(vlist[2], vlist[1], vlist[0]);
    x9, y9, z9 = _TYLV([None, None, None], [FlattenList([vlist[2], vlist[1], vlist[0]])])
    # (Line 77) ep_assert(tc2 + 1 == GetTriggerCounter(), "Wrongly elide var copy");
    ep_assert(tc2 + 1 == GetTriggerCounter(), "Wrongly elide var copy")
    # (Line 79) const tc3 = GetTriggerCounter();
    tc3 = GetTriggerCounter()
    # (Line 80) var x10 = dwread_epd(0);
    x10 = _TYLV([None], [f_dwread_epd(0)])
    # (Line 81) ep_assert(tc3 + 1 == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc3 + 1 == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 83) const foo = function () : TrgUnit { return 0; };
    @EUDTypedFunc([], [TrgUnit])
    def _lambda2():
        EUDReturn(0)

    foo = _lambda2
    # (Line 84) foo();  // create function body
    foo()
    # (Line 85) const tc4 = GetTriggerCounter();
    tc4 = GetTriggerCounter()
    # (Line 86) var baar = foo();
    baar = _TYLV([None], [foo()])
    # (Line 87) ep_assert(tc4 + 1 == GetTriggerCounter(), "Fail to elide var copy");
    ep_assert(tc4 + 1 == GetTriggerCounter(), "Fail to elide var copy")
    # (Line 89) cond.__lshift__(Memory(empty, AtLeast, 1));
    cond.__lshift__(Memory(empty, AtLeast, 1))
    # (Line 90) if (cond) { ret += 1 << 3; }
    if EUDIf()(cond):
        ret.__iadd__(_LSH(1,3))
        # (Line 91) ret += x;
    EUDEndIf()
    ret.__iadd__(x)
    # (Line 92) if(Is64BitWireframe()) {}
    if EUDIf()(Is64BitWireframe()):
        # (Line 93) var z = EUDVariable();
        pass
    EUDEndIf()
    z = _TYLV([None], [EUDVariable()])
    # (Line 94) math.cos(1);
    math.cos(1)
    # (Line 95) return ret;
    EUDReturn(ret)
    # (Line 96) }
